import typing as T

from google.protobuf.descriptor_pb2 import DescriptorProto, FileDescriptorProto, FileDescriptorSet


JsonType = T.Union[str, dict[str, T.Any]]


def process_proto_message(
    message: DescriptorProto, file_descriptor: FileDescriptorProto
) -> dict[str, T.Any]:
    schema: dict[str, T.Any] = {}
    for field in message.field:
        json_type: JsonType

        if field.type in (1, 2):
            json_type = "number"
        elif field.type in (3, 4, 5, 13, 17, 18):
            json_type = "integer"
        elif field.type == 9:
            json_type = "string"
        elif field.type == 11:
            json_type = "object"
        elif field.type == 14:
            json_type = "enum"
        else:
            json_type = "string"

        if field.label == 3:
            json_type = {"type": "array", "items": {"type": json_type}}

        if field.type == 11:
            nested_name = field.type_name.split(".")[-1]
            nested_message = next(
                (m for m in file_descriptor.message_type if m.name == nested_name),
                None,
            )
            schema[field.name] = (
                process_proto_message(nested_message, file_descriptor)
                if nested_message
                else {"type": "object"}
            )
        else:
            schema[field.name] = json_type

    return schema


def proto_to_schema(descriptor_file: str, message_name: str) -> dict[str, T.Any]:
    fds = FileDescriptorSet()
    with open(descriptor_file, "rb") as infile:
        fds.ParseFromString(infile.read())

    message = None
    file_descriptor = None
    for current_file_descriptor in fds.file:
        for msg in current_file_descriptor.message_type:
            if msg.name == message_name:
                message = msg
                file_descriptor = current_file_descriptor
                break
        if message is not None:
            break

    if message is None or file_descriptor is None:
        raise ValueError(f"Message {message_name} not found in descriptor file")

    return process_proto_message(message, file_descriptor)



"""Shared protobuf utilities."""

from .pb import get_enum_val_from_str, get_all_field_names, increment_timestamp, redact_fields
from .schema import proto_to_schema

__all__ = [
    "get_enum_val_from_str",
    "get_all_field_names",
    "increment_timestamp",
    "redact_fields",
    "proto_to_schema",
]



import typing as T

from google.protobuf.internal.enum_type_wrapper import EnumTypeWrapper
from google.protobuf.message import Message
from google.protobuf.timestamp_pb2 import Timestamp


def get_enum_val_from_str(raw: str, enum_type: T.Any) -> T.Optional[int]:
    value = raw.strip().upper()
    if not isinstance(enum_type, EnumTypeWrapper):
        return None
    enum_val = enum_type.DESCRIPTOR.values_by_name.get(value)
    return None if enum_val is None else int(enum_val.number)


def increment_timestamp(incremental_time_seconds: float, timestamp_pb: Timestamp) -> Timestamp:
    timestamp_pb.nanos = timestamp_pb.nanos + int(
        (incremental_time_seconds - int(incremental_time_seconds)) * 1e9
    )
    if timestamp_pb.nanos >= 1e9:
        timestamp_pb.seconds += 1
        timestamp_pb.nanos -= int(1e9)

    timestamp_pb.seconds += int(incremental_time_seconds)
    return timestamp_pb


def get_all_field_names(
    message: Message | None, all_fields: list[str], fields: list[str] | None = None
) -> None:
    if message is None:
        return

    for field in fields or list(message.DESCRIPTOR.fields_by_name.keys()):
        try:
            descriptor = message.DESCRIPTOR.fields_by_name[field]
            all_fields.append(field)
            if descriptor.message_type:
                field_value = getattr(message, field)
                if field_value:
                    nested: list[str] = []
                    get_all_field_names(field_value, nested)
                    for item in nested:
                        all_fields.append(f"{field}.{item}")
        except KeyError:
            pass


def redact_fields(message: Message, redacted_fields: dict[str, bool]) -> None:
    for field_path, should_redact in redacted_fields.items():
        if not should_redact:
            continue

        parts = field_path.split(".")
        current: T.Any = message
        for part in parts[:-1]:
            if hasattr(current, part):
                current = getattr(current, part)
            else:
                current = None
                break

        if current is None:
            continue

        final_field = parts[-1]
        if hasattr(current, final_field):
            current.ClearField(final_field)
